//CUBE texture mapping demo by Ivaylo Beltchev
//email: ivob@geocities.com
//url:   http://www.geocities.com/SiliconValley/Bay/2234

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

#include "grlib.h"
#ifdef __WINDOWS__
#include "resource.h"
#endif

#define PI 3.1415

#define VY 3

#define POINTN  8
#define SIDEN   6
#define BMPA  128
#define BMPB   10
#define BMPC    5

#define A 0.01
#define MAXA 0.10

double ax,ay,az;

struct Tpoint {
  double x,y,z;
  int    px,py;
  signed char a,b;
};

Tpoint point[POINTN];

struct Tside {
  int p1,p2,p3,p4;
};

class Tcubeapp: public TGRapp {
  public:
    Tcubeapp(char *_title): TGRapp(_title) {};
    virtual int init( void );
    virtual void tini( void );
    virtual int setmode(int width,int height,int bpp);
    virtual void draw( void );
    virtual void eventhandler(Tevent *ev);
#ifdef __WINDOWS__
    virtual int wndproc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam);
#endif
} cubeapp("GRlib Textured Cube");

Tside side[SIDEN];
unsigned char *texture[6];

int *lptr;
int *rptr;
int *lx;
int *rx;
signed char *la;
signed char *ra;
signed char *lb;
signed char *rb;
int *dy,*ddy;
int SCALE;
unsigned char *buffer8;
unsigned short *buffer16;
unsigned char *buffer24;
unsigned long *buffer32;
int shade;
int mapping=0; // 0-none, 1-flat, 2-goraud, 3-phong, 4-transparency, 5-inclined mapping
int bits;      // 0-8, 1-16, 2-24

Trgb col[256];

char phongtbl[256][256];
char paltbl8[256][256];
char trtbl8[256][256];
short paltbl16[256][256];
long paltbl32[256][256];
unsigned char trtbl1[256][256];
short trtbl2[256][256];
short col2[256];
long  col3[256];

void initcube( void )
{
  point[0].x=-1;point[0].y=-1+VY;point[0].z=-1;
  point[1].x= 1;point[1].y=-1+VY;point[1].z=-1;
  point[2].x= 1;point[2].y= 1+VY;point[2].z=-1;
  point[3].x=-1;point[3].y= 1+VY;point[3].z=-1;
  point[4].x=-1;point[4].y=-1+VY;point[4].z= 1;
  point[5].x= 1;point[5].y=-1+VY;point[5].z= 1;
  point[6].x= 1;point[6].y= 1+VY;point[6].z= 1;
  point[7].x=-1;point[7].y= 1+VY;point[7].z= 1;
  side[0].p1=0;side[0].p2=1;side[0].p3=2;side[0].p4=3;
  side[1].p1=0;side[1].p2=4;side[1].p3=5;side[1].p4=1;
  side[2].p1=1;side[2].p2=5;side[2].p3=6;side[2].p4=2;
  side[3].p1=2;side[3].p2=6;side[3].p3=7;side[3].p4=3;
  side[4].p1=3;side[4].p2=7;side[4].p3=4;side[4].p4=0;
  side[5].p1=4;side[5].p2=7;side[5].p3=6;side[5].p4=5;
}

int readtextures( void )
{
  FILE *f;
  char s[20];

  for (int i=0;i<6;i++) {
    sprintf(s,"T%d.BMP",i+1);
    f=fopen(s,"rb");
    if (f==NULL) return 0;
    fseek(f,0x436,SEEK_SET);
    texture[i]=(unsigned char *)malloc(BMPA*BMPA);
    fread(texture[i],1,BMPA*BMPA,f);
    fclose(f);
  }
  return 1;
}

void rotx(double a)
{
  double s=sin(a);
  double c=cos(a);

  for (int i=0;i<POINTN;i++) {
    double y=(point[i].y-VY)*c+point[i].z*s;
    double z=-(point[i].y-VY)*s+point[i].z*c;
    point[i].y=y+VY;
    point[i].z=z;
  }
}

void roty(double a)
{
  double s=sin(a);
  double c=cos(a);

  for (int i=0;i<POINTN;i++) {
    double x=point[i].x*c-point[i].z*s;
    double z=point[i].x*s+point[i].z*c;
    point[i].x=x;
    point[i].z=z;
  }
}

void rotz(double a)
{
  double s=sin(a);
  double c=cos(a);

  for (int i=0;i<POINTN;i++) {
    double x=point[i].x*c+(point[i].y-VY)*s;
    double y=-point[i].x*s+(point[i].y-VY)*c;
    point[i].x=x;
    point[i].y=y+VY;
  }
}

void project( void )
{
  for (int i=0;i<POINTN;i++) {
    double x=point[i].x;
    double y=point[i].y;
    double z=point[i].z;
    point[i].px=(int)(x*VY*SCALE/y+scr_ox);
    point[i].py=(int)(scr_oy-z*VY*SCALE/y);
    double a=atan2(y,x);
    double b=acos(z/sqrt(x*x+y*y+z*z));
    point[i].a=(char)((a/PI+1)*128);
    point[i].b=(char)((b/PI+0.5)*256);
  }
}

void setptr(int *ptr,double x1,int y1,double x2,int y2)
{
  double dx=(x2-x1)/(y2-y1);
  for (;y1<y2;y1++,x1+=dx)
    ptr[y1]=(int)x1;
}

void setptrp(signed char *ptr,double a1,int y1,double a2,int y2)
{
  double da=(a2-a1)/(double)(y2-y1);
  for (;y1<y2;y1++,a1+=da)
    ptr[y1]=(char)a1;
}

int miny,maxy;
double A1,A2,A3,B1,B2,B3,C1,C2,C3;
int qb;

void mapside8_noshading(int s)
{
  int x,y,xn;
  unsigned int a,b;
  unsigned da[2],db1[2];
  unsigned char *t=texture[s];
  unsigned char *bf;
  double A0,B0,C0;
  double cc=65536;

  for (y=miny;y<maxy;y++) {
    int v=scr_oy-y;
    int x0=lptr[y];
    xn=rptr[y]-x0;
    if (xn<=0) continue;
    int u=x0-scr_ox;
    C0=u*C1+v*C2+VY*SCALE*C3;
    double q=cc/C0;
    A0=u*A1+v*A2+VY*SCALE*A3;
    B0=u*B1+v*B2+VY*SCALE*B3;
    a=(int)(A0*q+BMPC*65536);
    b=(int)(B0*q+BMPC*65536);
    bf=buffer8+y*scr_cols+x0;
    double C2=C1*8;
    double DA=A1*C0-A0*C1;
    double DB1=B1*C0-B0*C1;
    _asm {
      fld     cc
      fld     DA
      fld     DB1
      fld     C2
      fld     C0
      fld     st
      fadd    st,st(2)
      fld     st
      fmul    st,st(2)
      fdivr   st,st(6)
    }
    for (x=0;x<xn;x++) {
      if ((x&7)==0) {
        _asm { 
          fld     st
          fmul    st,st(6)
          fistp   qword ptr da
          fmul    st,st(4)
          fistp   qword ptr db1
          fst     st(1)
          fadd    st,st(2)
          fld     st
          fmul    st,st(2)
          fdivr   st,st(6)
        }
      }
      unsigned int aa=a/65536;
      unsigned int bb=b/65536;
      aa%=BMPA;
      bb%=BMPA;
      *bf=*(t+aa*BMPA+bb);
      bf++;
      a+=da[0];
      b+=db1[0];
    }
    _asm {
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
    }
  }
}

void mapside8_flat(int s)
{
  int x,y,xn;
  unsigned int a,b;
  unsigned da[2],db1[2];
  unsigned char *t=texture[s];
  unsigned char *bf;
  char *pal;
  double A0,B0,C0;
  double cc=65536;

  pal=paltbl8[shade];
  for (y=miny;y<maxy;y++) {
    int v=scr_oy-y;
    int x0=lptr[y];
    xn=rptr[y]-x0;
    if (xn<=0) continue;
    int u=x0-scr_ox;
    C0=u*C1+v*C2+VY*SCALE*C3;
    double q=cc/C0;
    A0=u*A1+v*A2+VY*SCALE*A3;
    B0=u*B1+v*B2+VY*SCALE*B3;
    a=(int)(A0*q+BMPC*65536);
    b=(int)(B0*q+BMPC*65536);
    bf=buffer8+y*scr_cols+x0;
    double C2=C1*8;
    double DA=A1*C0-A0*C1;
    double DB1=B1*C0-B0*C1;
    _asm {
      fld     cc
      fld     DA
      fld     DB1
      fld     C2
      fld     C0
      fld     st
      fadd    st,st(2)
      fld     st
      fmul    st,st(2)
      fdivr   st,st(6)
    }
    for (x=0;x<xn;x++) {
      if ((x&7)==0) {
        _asm { 
          fld     st
          fmul    st,st(6)
          fistp   qword ptr da
          fmul    st,st(4)
          fistp   qword ptr db1
          fst     st(1)
          fadd    st,st(2)
          fld     st
          fmul    st,st(2)
          fdivr   st,st(6)
        }
      }
      unsigned int aa=a/65536;
      unsigned int bb=b/65536;
      aa%=BMPA;
      bb%=BMPA;
      *bf=pal[*(t+aa*BMPA+bb)];
      bf++;
      a+=da[0];
      b+=db1[0];
    }
    _asm {
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
    }
  }
}

void mapside8_goraud(int s)
{
  int x,y,xn;
  unsigned int a,b;
  unsigned da[2],db1[2];
  unsigned char *t=texture[s];
  unsigned char *bf;
  double A0,B0,C0;
  double cc=65536;

  for (y=miny;y<maxy;y++) {
    int v=scr_oy-y;
    int x0=lptr[y];
    xn=rptr[y]-x0;
    if (xn<=0) continue;
    int u=x0-scr_ox;
    C0=u*C1+v*C2+VY*SCALE*C3;
    double q=cc/C0;
    A0=u*A1+v*A2+VY*SCALE*A3;
    B0=u*B1+v*B2+VY*SCALE*B3;
    a=(int)(A0*q+BMPC*65536);
    b=(int)(B0*q+BMPC*65536);
    unsigned int sx=(128+lx[y])*65536;
    unsigned int dsx=(rx[y]-lx[y])*65536/xn;
    bf=buffer8+y*scr_cols+x0;
    double C2=C1*8;
    double DA=A1*C0-A0*C1;
    double DB1=B1*C0-B0*C1;
    _asm {
      fld     cc
      fld     DA
      fld     DB1
      fld     C2
      fld     C0
      fld     st
      fadd    st,st(2)
      fld     st
      fmul    st,st(2)
      fdivr   st,st(6)
    }
    for (x=0;x<xn;x++) {
      if ((x&7)==0) {
        _asm { 
          fld     st
          fmul    st,st(6)
          fistp   qword ptr da
          fmul    st,st(4)
          fistp   qword ptr db1
          fst     st(1)
          fadd    st,st(2)
          fld     st
          fmul    st,st(2)
          fdivr   st,st(6)
        }
      }
      unsigned int aa=a/65536;
      unsigned int bb=b/65536;
      aa%=BMPA;
      bb%=BMPA;
      unsigned int ssx=sx/65536;
      *bf=paltbl8[ssx][*(t+aa*BMPA+bb)];
      bf++;
      a+=da[0];
      b+=db1[0];
      sx+=dsx;
    }
    _asm {
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
    }
  }
}

void mapside8_phong(int s)
{
  int x,y,xn;
  unsigned int a,b;
  unsigned da[2],db1[2];
  unsigned char *t=texture[s];
  unsigned char *bf;
  double A0,B0,C0;
  static double cc=65536;

  for (y=miny;y<maxy;y++) {
    int v=scr_oy-y;
    int x0=lptr[y];
    xn=rptr[y]-x0;
    if (xn<=0) continue;
    int u=x0-scr_ox;
    C0=u*C1+v*C2+VY*SCALE*C3;
    double q=cc/C0;
    A0=u*A1+v*A2+VY*SCALE*A3;
    B0=u*B1+v*B2+VY*SCALE*B3;
    a=(int)(A0*q+BMPC*65536);
    b=(int)(B0*q+BMPC*65536);
    unsigned int sa=(int)(la[y]+128)*65536;
    unsigned int dsa=(int)(ra[y]-la[y])*65536/xn;
    unsigned int sb=(int)(lb[y]+128)*65536;
    unsigned int dsb=(int)(rb[y]-lb[y])*65536/xn;
    bf=buffer8+y*scr_cols+x0;
    double C2=C1*8;
    double DA=A1*C0-A0*C1;
    double DB1=B1*C0-B0*C1;
    _asm {
      fld     cc
      fld     DA
      fld     DB1
      fld     C2
      fld     C0
      fld     st
      fadd    st,st(2)
      fld     st
      fmul    st,st(2)
      fdivr   st,st(6)
    }
    for (x=0;x<xn;x++) {
      if ((x&7)==0) {
        _asm { 
          fld     st
          fmul    st,st(6)
          fistp   qword ptr da
          fmul    st,st(4)
          fistp   qword ptr db1
          fst     st(1)
          fadd    st,st(2)
          fld     st
          fmul    st,st(2)
          fdivr   st,st(6)
        }
      }
      unsigned char ssa=sa/65536;
      unsigned char ssb=sb/65536;
      unsigned char ss=phongtbl[ssa][ssb];
      unsigned int aa=a/65536;
      unsigned int bb=b/65536;
      aa%=BMPA;
      bb%=BMPA;
      *bf=paltbl8[ss][*(t+aa*BMPA+bb)];
      bf++;
      a+=da[0];
      b+=db1[0];
      sa+=dsa;
      sb+=dsb;
    }
    _asm {
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
    }
  }
}

void mapside8_transparence(int s)
{
  int x,y,xn;
  unsigned int a,b;
  unsigned da[2],db1[2];
  unsigned char *t=texture[s];
  unsigned char *bf;
  double A0,B0,C0;
  double cc=65536;

  for (y=miny;y<maxy;y++) {
    int v=scr_oy-y;
    int x0=lptr[y];
    xn=rptr[y]-x0;
    if (xn<=0) continue;
    int u=x0-scr_ox;
    C0=u*C1+v*C2+VY*SCALE*C3;
    double q=cc/C0;
    A0=u*A1+v*A2+VY*SCALE*A3;
    B0=u*B1+v*B2+VY*SCALE*B3;
    a=(int)(A0*q+BMPC*65536);
    b=(int)(B0*q+BMPC*65536);
    bf=buffer8+y*scr_cols+x0;
    double C2=C1*8;
    double DA=A1*C0-A0*C1;
    double DB1=B1*C0-B0*C1;
    _asm {
      fld     cc
      fld     DA
      fld     DB1
      fld     C2
      fld     C0
      fld     st
      fadd    st,st(2)
      fld     st
      fmul    st,st(2)
      fdivr   st,st(6)
    }
    for (x=0;x<xn;x++) {
      if ((x&7)==0) {
        _asm { 
          fld     st
          fmul    st,st(6)
          fistp   qword ptr da
          fmul    st,st(4)
          fistp   qword ptr db1
          fst     st(1)
          fadd    st,st(2)
          fld     st
          fmul    st,st(2)
          fdivr   st,st(6)
        }
      }
      unsigned int aa=a/65536;
      unsigned int bb=b/65536;
      aa%=BMPA;
      bb%=BMPA;
      *bf=trtbl8[*bf][*(t+aa*BMPA+bb)];
      bf++;
      a+=da[0];
      b+=db1[0];
    }
    _asm {
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
    }
  }
}

double DC;

void mapside8v(int s)
{
  int x,y;
  double cc=65536;
  unsigned int a,b,da,db;
  unsigned char *t=texture[s];
  unsigned char *bf;

  for (y=miny;y<maxy;y++) {
    int x0=lptr[y];
    int u=x0-scr_ox;
    int v=scr_oy-y-dy[x0];
    int xn=rptr[y];
    if (xn<=x0) continue;
    double C0=u*C1+v*C2+VY*SCALE*C3;
    double q=cc/C0;
    double A0=u*A1+v*A2+VY*SCALE*A3;
    double B0=u*B1+v*B2+VY*SCALE*B3;
    double DA=A1-A2*DC;
    double DB=B1-B2*DC;
    a=(unsigned int)(A0*q+BMPC*65536);
    b=(unsigned int)(B0*q+BMPC*65536);
    da=(unsigned int)(DA*q);
    db=(unsigned int)(DB*q);
    bf=buffer8+y*scr_cols;
    for (x=x0;x<xn;x++) {
      unsigned int aa=a/65536;
      unsigned int bb=b/65536;
      aa%=BMPA;
      bb%=BMPA;
      *(bf+ddy[x])=*(t+aa*BMPA+bb);
      a+=da;
      b+=db;
    }
  }
}

void mapside8h(int s)
{
  int x,y;
  double cc=65536;
  unsigned int a,b,da,db;
  unsigned char *t=texture[s];
  unsigned char *bf;

  for (x=miny;x<maxy;x++) {
    int y0=lptr[x];
    int u=x+dy[y0]-scr_ox;
    int v=scr_oy-y0;
    int yn=rptr[x];
    if (yn<=y0) continue;
    double C0=u*C1+v*C2+VY*SCALE*C3;
    double q=cc/C0;
    double A0=u*A1+v*A2+VY*SCALE*A3;
    double B0=u*B1+v*B2+VY*SCALE*B3;
    double DA=A1*DC-A2;
    double DB=B1*DC-B2;
    a=(unsigned int)(A0*q+BMPC*65536);
    b=(unsigned int)(B0*q+BMPC*65536);
    da=(unsigned int)(DA*q);
    db=(unsigned int)(DB*q);
    bf=buffer8+y0*scr_cols+x;
    for (y=y0;y<yn;y++) {
      unsigned int aa=a/65536;
      unsigned int bb=b/65536;
      aa%=BMPA;
      bb%=BMPA;
      *(bf+dy[y])=*(t+aa*BMPA+bb);
      bf+=scr_cols;
      a+=da;
      b+=db;
    }
  }
}

void drawsidei8(int s)
{
  Tpoint p[4];
  int *ptr1,*ptr2;

  p[0]=point[side[s].p1];
  p[1]=point[side[s].p2];
  p[2]=point[side[s].p3];
  p[3]=point[side[s].p4];
  double dx1=p[1].px-p[0].px;
  double dy1=p[1].py-p[0].py;
  double dx2=p[3].px-p[0].px;
  double dy2=p[3].py-p[0].py;
  if (dx1*dy2-dx2*dy1>0) {
    ptr1=rptr;
    ptr2=lptr;
  }
  else return;
  A1=p[3].z*p[0].y-p[3].y*p[0].z;
  A2=p[3].y*p[0].x-p[3].x*p[0].y;
  A3=p[3].x*p[0].z-p[3].z*p[0].x;

  B1=p[1].y*p[0].z-p[1].z*p[0].y;
  B2=p[1].x*p[0].y-p[1].y*p[0].x;
  B3=p[1].z*p[0].x-p[1].x*p[0].z;

  C1=(p[3].y*p[1].z-p[3].z*p[1].y+B1+A1)/(BMPA-BMPB);
  C2=(p[3].x*p[1].y-p[3].y*p[1].x+B2+A2)/(BMPA-BMPB);
  C3=(p[3].z*p[1].x-p[3].x*p[1].z+B3+A3)/(BMPA-BMPB);
  if (fabs(C1)>fabs(C2)) {
    miny=scr_cols;maxy=0;
    DC=C1?C2/C1:0;
    for (int y=-scr_oy;y<scr_oy;y++)
      dy[y+scr_oy]=(int)(y*DC);
    for (int i=0;i<4;i++) {
      int j=(i+1)&3;
      int xi=p[i].px-dy[p[i].py];
      int xj=p[j].px-dy[p[j].py];
      if (xi<xj) setptr(ptr2,p[i].py,xi,p[j].py,xj);
            else setptr(ptr1,p[j].py,xj,p[i].py,xi);
      if (xi<miny) miny=xi;
      if (xi>maxy) maxy=xi;
    }
    mapside8h(s);
  }
  else {
    miny=scr_rows;maxy=0;
    DC=C2?C1/C2:0;
    for (int x=-scr_ox;x<scr_ox;x++) {
      dy[x+scr_ox]=(int)(x*DC);
      ddy[x+scr_ox]=dy[x+scr_ox]*scr_cols+x+scr_ox;
    }
    for (int i=0;i<4;i++) {
      int j=(i+1)&3;
      int yi=p[i].py-dy[p[i].px];
      int yj=p[j].py-dy[p[j].px];
      if (yi<yj) setptr(ptr1,p[i].px,yi,p[j].px,yj);
            else setptr(ptr2,p[j].px,yj,p[i].px,yi);
      if (yi<miny) miny=yi;
      if (yi>maxy) maxy=yi;
    }
    mapside8v(s);
  }
}

void (*mapfunc8[])(int s)={mapside8_noshading,mapside8_flat,
                          mapside8_goraud,mapside8_phong,
                          mapside8_transparence};

void drawside8(int s)
{
  Tpoint p[4];
  int *ptr1,*ptr2,*sx1,*sx2;
  signed char *sa1,*sa2,*sb1,*sb2;

  if (mapping==5) {
    drawsidei8(s);
    return;
  }
  p[0]=point[side[s].p1];
  p[1]=point[side[s].p2];
  p[2]=point[side[s].p3];
  p[3]=point[side[s].p4];
  double dx1=p[1].px-p[0].px;
  double dy1=p[1].py-p[0].py;
  double dx2=p[3].px-p[0].px;
  double dy2=p[3].py-p[0].py;
  if (p[0].px==0) buffer8[1000]=255;
  if (dx1*dy2-dx2*dy1>0) {
    ptr1=rptr;
    sx1=rx;
    sa1=ra;
    sb1=rb;
    ptr2=lptr;
    sx2=lx;
    sa2=la;
    sb2=lb;
  }
  else {
    if (mapping!=4) return;
    ptr1=lptr;
    sx1=lx;
    sa1=la;
    sb1=lb;
    ptr2=rptr;
    sx2=rx;
    sa2=ra;
    sb2=rb;
  }
  A1=p[3].z*p[0].y-p[3].y*p[0].z;
  A2=p[3].y*p[0].x-p[3].x*p[0].y;
  A3=p[3].x*p[0].z-p[3].z*p[0].x;

  B1=p[1].y*p[0].z-p[1].z*p[0].y;
  B2=p[1].x*p[0].y-p[1].y*p[0].x;
  B3=p[1].z*p[0].x-p[1].x*p[0].z;

  C1=(p[3].y*p[1].z-p[3].z*p[1].y+B1+A1)/(BMPA-BMPB);
  C2=(p[3].x*p[1].y-p[3].y*p[1].x+B2+A2)/(BMPA-BMPB);
  C3=(p[3].z*p[1].x-p[3].x*p[1].z+B3+A3)/(BMPA-BMPB);
  miny=scr_rows;maxy=0;
  for (int i=0;i<4;i++) {
    int j=(i+1)&3;
    if (p[i].py<p[j].py) {
      setptr(ptr1,p[i].px,p[i].py,p[j].px,p[j].py);
      if (mapping==2) setptr(sx1,-73*p[i].x,p[i].py,-73*p[j].x,p[j].py); 
      if (mapping==3) {
        setptrp(sa1,p[i].a,p[i].py,p[j].a,p[j].py);
        setptrp(sb1,p[i].b,p[i].py,p[j].b,p[j].py);
      }
    }
    else {
      setptr(ptr2,p[j].px,p[j].py,p[i].px,p[i].py);
      if (mapping==2) setptr(sx2,-73*p[j].x,p[j].py,-73*p[i].x,p[i].py);
      if (mapping==3) {
        setptrp(sa2,p[j].a,p[j].py,p[i].a,p[i].py);
        setptrp(sb2,p[j].b,p[j].py,p[i].b,p[i].py);
      }
    }
    if (p[i].py<miny) miny=p[i].py;
    if (p[i].py>maxy) maxy=p[i].py;
  }
  if (mapping==1) shade=(int)(128-(p[0].x+p[1].x+p[2].x+p[3].x)*32);
  mapfunc8[mapping](s);
}








void mapside16_noshading(int s)
{
  int x,y,xn;
  unsigned int a,b;
  unsigned da[2],db1[2];
  unsigned char *t=texture[s];
  unsigned short *bf;
  double A0,B0,C0;
  double cc=65536;

  for (y=miny;y<maxy;y++) {
    int v=scr_oy-y;
    int x0=lptr[y];
    xn=rptr[y]-x0;
    if (xn<=0) continue;
    int u=x0-scr_ox;
    C0=u*C1+v*C2+VY*SCALE*C3;
    double q=cc/C0;
    A0=u*A1+v*A2+VY*SCALE*A3;
    B0=u*B1+v*B2+VY*SCALE*B3;
    a=(int)(A0*q+BMPC*65536);
    b=(int)(B0*q+BMPC*65536);
    bf=buffer16+y*scr_cols+x0;
    double C2=C1*8;
    double DA=A1*C0-A0*C1;
    double DB1=B1*C0-B0*C1;
    _asm {
      fld     cc
      fld     DA
      fld     DB1
      fld     C2
      fld     C0
      fld     st
      fadd    st,st(2)
      fld     st
      fmul    st,st(2)
      fdivr   st,st(6)
    }
    for (x=0;x<xn;x++) {
      if ((x&7)==0) {
        _asm { 
          fld     st
          fmul    st,st(6)
          fistp   qword ptr da
          fmul    st,st(4)
          fistp   qword ptr db1
          fst     st(1)
          fadd    st,st(2)
          fld     st
          fmul    st,st(2)
          fdivr   st,st(6)
        }
      }
      unsigned int aa=a/65536;
      unsigned int bb=b/65536;
      aa%=BMPA;
      bb%=BMPA;
      *bf=col2[*(t+aa*BMPA+bb)];
      bf++;
      a+=da[0];
      b+=db1[0];
    }
    _asm {
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
    }
  }
}

void mapside16_flat(int s)
{
  int x,y,xn;
  unsigned int a,b;
  unsigned da[2],db1[2];
  unsigned char *t=texture[s];
  unsigned short *bf;
  short *pal;
  double A0,B0,C0;
  double cc=65536;

  pal=paltbl16[shade];
  for (y=miny;y<maxy;y++) {
    int v=scr_oy-y;
    int x0=lptr[y];
    xn=rptr[y]-x0;
    if (xn<=0) continue;
    int u=x0-scr_ox;
    C0=u*C1+v*C2+VY*SCALE*C3;
    double q=cc/C0;
    A0=u*A1+v*A2+VY*SCALE*A3;
    B0=u*B1+v*B2+VY*SCALE*B3;
    a=(int)(A0*q+BMPC*65536);
    b=(int)(B0*q+BMPC*65536);
    bf=buffer16+y*scr_cols+x0;
    double C2=C1*8;
    double DA=A1*C0-A0*C1;
    double DB1=B1*C0-B0*C1;
    _asm {
      fld     cc
      fld     DA
      fld     DB1
      fld     C2
      fld     C0
      fld     st
      fadd    st,st(2)
      fld     st
      fmul    st,st(2)
      fdivr   st,st(6)
    }
    for (x=0;x<xn;x++) {
      if ((x&7)==0) {
        _asm { 
          fld     st
          fmul    st,st(6)
          fistp   qword ptr da
          fmul    st,st(4)
          fistp   qword ptr db1
          fst     st(1)
          fadd    st,st(2)
          fld     st
          fmul    st,st(2)
          fdivr   st,st(6)
        }
      }
      unsigned int aa=a/65536;
      unsigned int bb=b/65536;
      aa%=BMPA;
      bb%=BMPA;
      *bf=pal[*(t+aa*BMPA+bb)];
      bf++;
      a+=da[0];
      b+=db1[0];
    }
    _asm {
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
    }
  }
}

void mapside16_goraud(int s)
{
  int x,y,xn;
  unsigned int a,b;
  unsigned da[2],db1[2];
  unsigned char *t=texture[s];
  unsigned short *bf;
  double A0,B0,C0;
  double cc=65536;

  for (y=miny;y<maxy;y++) {
    int v=scr_oy-y;
    int x0=lptr[y];
    xn=rptr[y]-x0;
    if (xn<=0) continue;
    int u=x0-scr_ox;
    C0=u*C1+v*C2+VY*SCALE*C3;
    double q=cc/C0;
    A0=u*A1+v*A2+VY*SCALE*A3;
    B0=u*B1+v*B2+VY*SCALE*B3;
    a=(int)(A0*q+BMPC*65536);
    b=(int)(B0*q+BMPC*65536);
    unsigned int sx=(128+lx[y])*65536;
    unsigned int dsx=(rx[y]-lx[y])*65536/xn;
    bf=buffer16+y*scr_cols+x0;
    double C2=C1*8;
    double DA=A1*C0-A0*C1;
    double DB1=B1*C0-B0*C1;
    _asm {
      fld     cc
      fld     DA
      fld     DB1
      fld     C2
      fld     C0
      fld     st
      fadd    st,st(2)
      fld     st
      fmul    st,st(2)
      fdivr   st,st(6)
    }
    for (x=0;x<xn;x++) {
      if ((x&7)==0) {
        _asm { 
          fld     st
          fmul    st,st(6)
          fistp   qword ptr da
          fmul    st,st(4)
          fistp   qword ptr db1
          fst     st(1)
          fadd    st,st(2)
          fld     st
          fmul    st,st(2)
          fdivr   st,st(6)
        }
      }
      unsigned int aa=a/65536;
      unsigned int bb=b/65536;
      aa%=BMPA;
      bb%=BMPA;
      unsigned int ssx=sx/65536;
      *bf=paltbl16[ssx][*(t+aa*BMPA+bb)];
      bf++;
      a+=da[0];
      b+=db1[0];
      sx+=dsx;
    }
    _asm {
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
    }
  }
}

void mapside16_phong(int s)
{
  int x,y,xn;
  unsigned int a,b;
  unsigned da[2],db1[2];
  unsigned char *t=texture[s];
  unsigned short *bf;
  double A0,B0,C0;
  static double cc=65536;

  for (y=miny;y<maxy;y++) {
    int v=scr_oy-y;
    int x0=lptr[y];
    xn=rptr[y]-x0;
    if (xn<=0) continue;
    int u=x0-scr_ox;
    C0=u*C1+v*C2+VY*SCALE*C3;
    double q=cc/C0;
    A0=u*A1+v*A2+VY*SCALE*A3;
    B0=u*B1+v*B2+VY*SCALE*B3;
    a=(int)(A0*q+BMPC*65536);
    b=(int)(B0*q+BMPC*65536);
    unsigned int sa=(int)(la[y]+128)*65536;
    unsigned int dsa=(int)(ra[y]-la[y])*65536/xn;
    unsigned int sb=(int)(lb[y]+128)*65536;
    unsigned int dsb=(int)(rb[y]-lb[y])*65536/xn;
    bf=buffer16+y*scr_cols+x0;
    double C2=C1*8;
    double DA=A1*C0-A0*C1;
    double DB1=B1*C0-B0*C1;
    _asm {
      fld     cc
      fld     DA
      fld     DB1
      fld     C2
      fld     C0
      fld     st
      fadd    st,st(2)
      fld     st
      fmul    st,st(2)
      fdivr   st,st(6)
    }
    for (x=0;x<xn;x++) {
      if ((x&7)==0) {
        _asm { 
          fld     st
          fmul    st,st(6)
          fistp   qword ptr da
          fmul    st,st(4)
          fistp   qword ptr db1
          fst     st(1)
          fadd    st,st(2)
          fld     st
          fmul    st,st(2)
          fdivr   st,st(6)
        }
      }
      unsigned char ssa=sa/65536;
      unsigned char ssb=sb/65536;
      unsigned char ss=phongtbl[ssa][ssb];
      unsigned int aa=a/65536;
      unsigned int bb=b/65536;
      aa%=BMPA;
      bb%=BMPA;
      *bf=paltbl16[ss][*(t+aa*BMPA+bb)];
      bf++;
      a+=da[0];
      b+=db1[0];
      sa+=dsa;
      sb+=dsb;
    }
    _asm {
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
    }
  }
}

void mapside16_transparence(int s)
{
  int x,y,xn;
  unsigned int a,b;
  unsigned da[2],db1[2];
  unsigned char *t=texture[s];
  unsigned short *bf;
  double A0,B0,C0;
  double cc=65536;

  for (y=miny;y<maxy;y++) {
    int v=scr_oy-y;
    int x0=lptr[y];
    xn=rptr[y]-x0;
    if (xn<=0) continue;
    int u=x0-scr_ox;
    C0=u*C1+v*C2+VY*SCALE*C3;
    double q=cc/C0;
    A0=u*A1+v*A2+VY*SCALE*A3;
    B0=u*B1+v*B2+VY*SCALE*B3;
    a=(int)(A0*q+BMPC*65536);
    b=(int)(B0*q+BMPC*65536);
    bf=buffer16+y*scr_cols+x0;
    double C2=C1*8;
    double DA=A1*C0-A0*C1;
    double DB1=B1*C0-B0*C1;
    _asm {
      fld     cc
      fld     DA
      fld     DB1
      fld     C2
      fld     C0
      fld     st
      fadd    st,st(2)
      fld     st
      fmul    st,st(2)
      fdivr   st,st(6)
    }
    for (x=0;x<xn;x++) {
      if ((x&7)==0) {
        _asm { 
          fld     st
          fmul    st,st(6)
          fistp   qword ptr da
          fmul    st,st(4)
          fistp   qword ptr db1
          fst     st(1)
          fadd    st,st(2)
          fld     st
          fmul    st,st(2)
          fdivr   st,st(6)
        }
      }
      unsigned int aa=a/65536;
      unsigned int bb=b/65536;
      aa%=BMPA;
      bb%=BMPA;
      unsigned char ct=*(t+aa*BMPA+bb);
      *bf=trtbl1[((unsigned char *)bf)[0]][ct]+
          trtbl2[((unsigned char *)bf)[1]][ct];
      bf++;
      a+=da[0];
      b+=db1[0];
    }
    _asm {
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
    }
  }
}

void mapside16v(int s)
{
  int x,y;
  double cc=65536;
  unsigned int a,b,da,db;
  unsigned char *t=texture[s];
  unsigned short *bf;

  for (y=miny;y<maxy;y++) {
    int x0=lptr[y];
    int u=x0-scr_ox;
    int v=scr_oy-y-dy[x0];
    int xn=rptr[y];
    if (xn<=x0) continue;
    double C0=u*C1+v*C2+VY*SCALE*C3;
    double q=cc/C0;
    double A0=u*A1+v*A2+VY*SCALE*A3;
    double B0=u*B1+v*B2+VY*SCALE*B3;
    double DA=A1-A2*DC;
    double DB=B1-B2*DC;
    a=(unsigned int)(A0*q+BMPC*65536);
    b=(unsigned int)(B0*q+BMPC*65536);
    da=(unsigned int)(DA*q);
    db=(unsigned int)(DB*q);
    bf=buffer16+y*scr_cols;
    for (x=x0;x<xn;x++) {
      unsigned int aa=a/65536;
      unsigned int bb=b/65536;
      aa%=BMPA;
      bb%=BMPA;
      *(bf+ddy[x])=col2[*(t+aa*BMPA+bb)];
      a+=da;
      b+=db;
    }
  }
}

void mapside16h(int s)
{
  int x,y;
  double cc=65536;
  unsigned int a,b,da,db;
  unsigned char *t=texture[s];
  unsigned short *bf;

  for (x=miny;x<maxy;x++) {
    int y0=lptr[x];
    int u=x+dy[y0]-scr_ox;
    int v=scr_oy-y0;
    int yn=rptr[x];
    if (yn<=y0) continue;
    double C0=u*C1+v*C2+VY*SCALE*C3;
    double q=cc/C0;
    double A0=u*A1+v*A2+VY*SCALE*A3;
    double B0=u*B1+v*B2+VY*SCALE*B3;
    double DA=A1*DC-A2;
    double DB=B1*DC-B2;
    a=(unsigned int)(A0*q+BMPC*65536);
    b=(unsigned int)(B0*q+BMPC*65536);
    da=(unsigned int)(DA*q);
    db=(unsigned int)(DB*q);
    bf=buffer16+y0*scr_cols+x;
    for (y=y0;y<yn;y++) {
      unsigned int aa=a/65536;
      unsigned int bb=b/65536;
      aa%=BMPA;
      bb%=BMPA;
      *(bf+dy[y])=col2[*(t+aa*BMPA+bb)];
      bf+=scr_cols;
      a+=da;
      b+=db;
    }
  }
}

void drawsidei16(int s)
{
  Tpoint p[4];
  int *ptr1,*ptr2;

  p[0]=point[side[s].p1];
  p[1]=point[side[s].p2];
  p[2]=point[side[s].p3];
  p[3]=point[side[s].p4];
  double dx1=p[1].px-p[0].px;
  double dy1=p[1].py-p[0].py;
  double dx2=p[3].px-p[0].px;
  double dy2=p[3].py-p[0].py;
  if (dx1*dy2-dx2*dy1>0) {
    ptr1=rptr;
    ptr2=lptr;
  }
  else return;
  A1=p[3].z*p[0].y-p[3].y*p[0].z;
  A2=p[3].y*p[0].x-p[3].x*p[0].y;
  A3=p[3].x*p[0].z-p[3].z*p[0].x;

  B1=p[1].y*p[0].z-p[1].z*p[0].y;
  B2=p[1].x*p[0].y-p[1].y*p[0].x;
  B3=p[1].z*p[0].x-p[1].x*p[0].z;

  C1=(p[3].y*p[1].z-p[3].z*p[1].y+B1+A1)/(BMPA-BMPB);
  C2=(p[3].x*p[1].y-p[3].y*p[1].x+B2+A2)/(BMPA-BMPB);
  C3=(p[3].z*p[1].x-p[3].x*p[1].z+B3+A3)/(BMPA-BMPB);
  if (fabs(C1)>fabs(C2)) {
    miny=scr_cols;maxy=0;
    DC=C1?C2/C1:0;
    for (int y=-scr_oy;y<scr_oy;y++)
      dy[y+scr_oy]=(int)(y*DC);
    for (int i=0;i<4;i++) {
      int j=(i+1)&3;
      int xi=p[i].px-dy[p[i].py];
      int xj=p[j].px-dy[p[j].py];
      if (xi<xj) setptr(ptr2,p[i].py,xi,p[j].py,xj);
            else setptr(ptr1,p[j].py,xj,p[i].py,xi);
      if (xi<miny) miny=xi;
      if (xi>maxy) maxy=xi;
    }
    mapside16h(s);
  }
  else {
    miny=scr_rows;maxy=0;
    DC=C2?C1/C2:0;
    for (int x=-scr_ox;x<scr_ox;x++) {
      dy[x+scr_ox]=(int)(x*DC);
      ddy[x+scr_ox]=dy[x+scr_ox]*scr_cols+x+scr_ox;
    }
    for (int i=0;i<4;i++) {
      int j=(i+1)&3;
      int yi=p[i].py-dy[p[i].px];
      int yj=p[j].py-dy[p[j].px];
      if (yi<yj) setptr(ptr1,p[i].px,yi,p[j].px,yj);
            else setptr(ptr2,p[j].px,yj,p[i].px,yi);
      if (yi<miny) miny=yi;
      if (yi>maxy) maxy=yi;
    }
    mapside16v(s);
  }
}

void (*mapfunc16[])(int s)={mapside16_noshading,mapside16_flat,
                          mapside16_goraud,mapside16_phong,
                          mapside16_transparence};

void drawside16(int s)
{
  Tpoint p[4];
  int *ptr1,*ptr2,*sx1,*sx2;
  signed char *sa1,*sa2,*sb1,*sb2;

  if (mapping==5) {
    drawsidei16(s);
    return;
  }
  p[0]=point[side[s].p1];
  p[1]=point[side[s].p2];
  p[2]=point[side[s].p3];
  p[3]=point[side[s].p4];
  double dx1=p[1].px-p[0].px;
  double dy1=p[1].py-p[0].py;
  double dx2=p[3].px-p[0].px;
  double dy2=p[3].py-p[0].py;
  if (dx1*dy2-dx2*dy1>0) {
    ptr1=rptr;
    sx1=rx;
    sa1=ra;
    sb1=rb;
    ptr2=lptr;
    sx2=lx;
    sa2=la;
    sb2=lb;
  }
  else {
    if (mapping!=4) return;
    ptr1=lptr;
    sx1=lx;
    sa1=la;
    sb1=lb;
    ptr2=rptr;
    sx2=rx;
    sa2=ra;
    sb2=rb;
  }
  A1=p[3].z*p[0].y-p[3].y*p[0].z;
  A2=p[3].y*p[0].x-p[3].x*p[0].y;
  A3=p[3].x*p[0].z-p[3].z*p[0].x;

  B1=p[1].y*p[0].z-p[1].z*p[0].y;
  B2=p[1].x*p[0].y-p[1].y*p[0].x;
  B3=p[1].z*p[0].x-p[1].x*p[0].z;

  C1=(p[3].y*p[1].z-p[3].z*p[1].y+B1+A1)/(BMPA-BMPB);
  C2=(p[3].x*p[1].y-p[3].y*p[1].x+B2+A2)/(BMPA-BMPB);
  C3=(p[3].z*p[1].x-p[3].x*p[1].z+B3+A3)/(BMPA-BMPB);
  miny=scr_rows;maxy=0;
  for (int i=0;i<4;i++) {
    int j=(i+1)&3;
    if (p[i].py<p[j].py) {
      setptr(ptr1,p[i].px,p[i].py,p[j].px,p[j].py);
      if (mapping==2) setptr(sx1,-73*p[i].x,p[i].py,-73*p[j].x,p[j].py); 
      if (mapping==3) {
        setptrp(sa1,p[i].a,p[i].py,p[j].a,p[j].py); 
        setptrp(sb1,p[i].b,p[i].py,p[j].b,p[j].py); 
      }
    }
    else {
      setptr(ptr2,p[j].px,p[j].py,p[i].px,p[i].py);
      if (mapping==2) setptr(sx2,-73*p[j].x,p[j].py,-73*p[i].x,p[i].py);
      if (mapping==3) {
        setptrp(sa2,p[j].a,p[j].py,p[i].a,p[i].py);
        setptrp(sb2,p[j].b,p[j].py,p[i].b,p[i].py);
      }
    }
    if (p[i].py<miny) miny=p[i].py;
    if (p[i].py>maxy) maxy=p[i].py;
  }
  if (mapping==1) shade=(int)(128-(p[0].x+p[1].x+p[2].x+p[3].x)*32);
  mapfunc16[mapping](s);
}









void mapside24_noshading(int s)
{
  int x,y,xn;
  unsigned int a,b;
  unsigned da[2],db1[2];
  unsigned char *t=texture[s];
  unsigned char *bf;
  double A0,B0,C0;
  double cc=65536;

  for (y=miny;y<maxy;y++) {
    int v=scr_oy-y;
    int x0=lptr[y];
    xn=rptr[y]-x0;
    if (xn<=0) continue;
    int u=x0-scr_ox;
    C0=u*C1+v*C2+VY*SCALE*C3;
    double q=cc/C0;
    A0=u*A1+v*A2+VY*SCALE*A3;
    B0=u*B1+v*B2+VY*SCALE*B3;
    a=(int)(A0*q+BMPC*65536);
    b=(int)(B0*q+BMPC*65536);
    bf=buffer24+(y*scr_cols+x0)*3;
    double C2=C1*8;
    double DA=A1*C0-A0*C1;
    double DB1=B1*C0-B0*C1;
    _asm {
      fld     cc
      fld     DA
      fld     DB1
      fld     C2
      fld     C0
      fld     st
      fadd    st,st(2)
      fld     st
      fmul    st,st(2)
      fdivr   st,st(6)
    }
    for (x=0;x<xn;x++) {
      if ((x&7)==0) {
        _asm { 
          fld     st
          fmul    st,st(6)
          fistp   qword ptr da
          fmul    st,st(4)
          fistp   qword ptr db1
          fst     st(1)
          fadd    st,st(2)
          fld     st
          fmul    st,st(2)
          fdivr   st,st(6)
        }
      }
      unsigned int aa=a/65536;
      unsigned int bb=b/65536;
      aa%=BMPA;
      bb%=BMPA;
      long c=col3[*(t+aa*BMPA+bb)];
      *(short *)bf=(short)c;
      *(bf+2)=(char)(c>>16);
      bf+=3;
      a+=da[0];
      b+=db1[0];
    }
    _asm {
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
    }
  }
}

void mapside24_flat(int s)
{
  int x,y,xn;
  unsigned int a,b;
  unsigned da[2],db1[2];
  unsigned char *t=texture[s];
  unsigned char *bf;
  long *pal;
  double A0,B0,C0;
  double cc=65536;

  pal=paltbl32[shade];
  for (y=miny;y<maxy;y++) {
    int v=scr_oy-y;
    int x0=lptr[y];
    xn=rptr[y]-x0;
    if (xn<=0) continue;
    int u=x0-scr_ox;
    C0=u*C1+v*C2+VY*SCALE*C3;
    double q=cc/C0;
    A0=u*A1+v*A2+VY*SCALE*A3;
    B0=u*B1+v*B2+VY*SCALE*B3;
    a=(int)(A0*q+BMPC*65536);
    b=(int)(B0*q+BMPC*65536);
    bf=buffer24+(y*scr_cols+x0)*3;
    double C2=C1*8;
    double DA=A1*C0-A0*C1;
    double DB1=B1*C0-B0*C1;
    _asm {
      fld     cc
      fld     DA
      fld     DB1
      fld     C2
      fld     C0
      fld     st
      fadd    st,st(2)
      fld     st
      fmul    st,st(2)
      fdivr   st,st(6)
    }
    for (x=0;x<xn;x++) {
      if ((x&7)==0) {
        _asm { 
          fld     st
          fmul    st,st(6)
          fistp   qword ptr da
          fmul    st,st(4)
          fistp   qword ptr db1
          fst     st(1)
          fadd    st,st(2)
          fld     st
          fmul    st,st(2)
          fdivr   st,st(6)
        }
      }
      unsigned int aa=a/65536;
      unsigned int bb=b/65536;
      aa%=BMPA;
      bb%=BMPA;
      long c=pal[*(t+aa*BMPA+bb)];
      *(short *)bf=(short)c;
      *(bf+2)=(char)(c>>16);
      bf+=3;
      a+=da[0];
      b+=db1[0];
    }
    _asm {
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
    }
  }
}

void mapside24_goraud(int s)
{
  int x,y,xn;
  unsigned int a,b;
  unsigned da[2],db1[2];
  unsigned char *t=texture[s];
  unsigned char *bf;
  double A0,B0,C0;
  double cc=65536;

  for (y=miny;y<maxy;y++) {
    int v=scr_oy-y;
    int x0=lptr[y];
    xn=rptr[y]-x0;
    if (xn<=0) continue;
    int u=x0-scr_ox;
    C0=u*C1+v*C2+VY*SCALE*C3;
    double q=cc/C0;
    A0=u*A1+v*A2+VY*SCALE*A3;
    B0=u*B1+v*B2+VY*SCALE*B3;
    a=(int)(A0*q+BMPC*65536);
    b=(int)(B0*q+BMPC*65536);
    unsigned int sx=(128+lx[y])*65536;
    unsigned int dsx=(rx[y]-lx[y])*65536/xn;
    bf=buffer24+(y*scr_cols+x0)*3;
    double C2=C1*8;
    double DA=A1*C0-A0*C1;
    double DB1=B1*C0-B0*C1;
    _asm {
      fld     cc
      fld     DA
      fld     DB1
      fld     C2
      fld     C0
      fld     st
      fadd    st,st(2)
      fld     st
      fmul    st,st(2)
      fdivr   st,st(6)
    }
    for (x=0;x<xn;x++) {
      if ((x&7)==0) {
        _asm { 
          fld     st
          fmul    st,st(6)
          fistp   qword ptr da
          fmul    st,st(4)
          fistp   qword ptr db1
          fst     st(1)
          fadd    st,st(2)
          fld     st
          fmul    st,st(2)
          fdivr   st,st(6)
        }
      }
      unsigned int aa=a/65536;
      unsigned int bb=b/65536;
      aa%=BMPA;
      bb%=BMPA;
      unsigned int ssx=sx/65536;
      long c=paltbl32[ssx][*(t+aa*BMPA+bb)];
      *(short *)bf=(short)c;
      *(bf+2)=(char)(c>>16);
      bf+=3;
      a+=da[0];
      b+=db1[0];
      sx+=dsx;
    }
    _asm {
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
    }
  }
}

void mapside24_phong(int s)
{
  int x,y,xn;
  unsigned int a,b;
  unsigned da[2],db1[2];
  unsigned char *t=texture[s];
  unsigned char *bf;
  double A0,B0,C0;
  static double cc=65536;

  for (y=miny;y<maxy;y++) {
    int v=scr_oy-y;
    int x0=lptr[y];
    xn=rptr[y]-x0;
    if (xn<=0) continue;
    int u=x0-scr_ox;
    C0=u*C1+v*C2+VY*SCALE*C3;
    double q=cc/C0;
    A0=u*A1+v*A2+VY*SCALE*A3;
    B0=u*B1+v*B2+VY*SCALE*B3;
    a=(int)(A0*q+BMPC*65536);
    b=(int)(B0*q+BMPC*65536);
    unsigned int sa=(int)(la[y]+128)*65536;
    unsigned int dsa=(int)(ra[y]-la[y])*65536/xn;
    unsigned int sb=(int)(lb[y]+128)*65536;
    unsigned int dsb=(int)(rb[y]-lb[y])*65536/xn;
    bf=buffer24+(y*scr_cols+x0)*3;
    double C2=C1*8;
    double DA=A1*C0-A0*C1;
    double DB1=B1*C0-B0*C1;
    _asm {
      fld     cc
      fld     DA
      fld     DB1
      fld     C2
      fld     C0
      fld     st
      fadd    st,st(2)
      fld     st
      fmul    st,st(2)
      fdivr   st,st(6)
    }
    for (x=0;x<xn;x++) {
      if ((x&7)==0) {
        _asm { 
          fld     st
          fmul    st,st(6)
          fistp   qword ptr da
          fmul    st,st(4)
          fistp   qword ptr db1
          fst     st(1)
          fadd    st,st(2)
          fld     st
          fmul    st,st(2)
          fdivr   st,st(6)
        }
      }
      unsigned char ssa=sa/65536;
      unsigned char ssb=sb/65536;
      unsigned char ss=phongtbl[ssa][ssb];
      unsigned int aa=a/65536;
      unsigned int bb=b/65536;
      aa%=BMPA;
      bb%=BMPA;
      long c=paltbl32[ss][*(t+aa*BMPA+bb)];
      *(short *)bf=(short)c;
      *(bf+2)=(char)(c>>16);
      bf+=3;
      a+=da[0];
      b+=db1[0];
      sa+=dsa;
      sb+=dsb;
    }
    _asm {
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
    }
  }
}

void (*mapfunc24[])(int s)={mapside24_noshading,mapside24_flat,
                          mapside24_goraud,mapside24_phong};

void drawside24(int s)
{
  Tpoint p[4];
  int *ptr1,*ptr2,*sx1,*sx2;
  signed char *sa1,*sa2,*sb1,*sb2;

  if (mapping>3) return;
  p[0]=point[side[s].p1];
  p[1]=point[side[s].p2];
  p[2]=point[side[s].p3];
  p[3]=point[side[s].p4];
  double dx1=p[1].px-p[0].px;
  double dy1=p[1].py-p[0].py;
  double dx2=p[3].px-p[0].px;
  double dy2=p[3].py-p[0].py;
  if (dx1*dy2-dx2*dy1>0) {
    ptr1=rptr;
    sx1=rx;
    sa1=ra;
    sb1=rb;
    ptr2=lptr;
    sx2=lx;
    sa2=la;
    sb2=lb;
  }
  else return;
  A1=p[3].z*p[0].y-p[3].y*p[0].z;
  A2=p[3].y*p[0].x-p[3].x*p[0].y;
  A3=p[3].x*p[0].z-p[3].z*p[0].x;

  B1=p[1].y*p[0].z-p[1].z*p[0].y;
  B2=p[1].x*p[0].y-p[1].y*p[0].x;
  B3=p[1].z*p[0].x-p[1].x*p[0].z;

  C1=(p[3].y*p[1].z-p[3].z*p[1].y+B1+A1)/(BMPA-BMPB);
  C2=(p[3].x*p[1].y-p[3].y*p[1].x+B2+A2)/(BMPA-BMPB);
  C3=(p[3].z*p[1].x-p[3].x*p[1].z+B3+A3)/(BMPA-BMPB);
  miny=scr_rows;maxy=0;
  for (int i=0;i<4;i++) {
    int j=(i+1)&3;
    if (p[i].py<p[j].py) {
      setptr(ptr1,p[i].px,p[i].py,p[j].px,p[j].py);
      if (mapping==2) setptr(sx1,-73*p[i].x,p[i].py,-73*p[j].x,p[j].py); 
      if (mapping==3) {
        setptrp(sa1,p[i].a,p[i].py,p[j].a,p[j].py); 
        setptrp(sb1,p[i].b,p[i].py,p[j].b,p[j].py); 
      }
    }
    else {
      setptr(ptr2,p[j].px,p[j].py,p[i].px,p[i].py);
      if (mapping==2) setptr(sx2,-73*p[j].x,p[j].py,-73*p[i].x,p[i].py);
      if (mapping==3) {
        setptrp(sa2,p[j].a,p[j].py,p[i].a,p[i].py);
        setptrp(sb2,p[j].b,p[j].py,p[i].b,p[i].py);
      }
    }
    if (p[i].py<miny) miny=p[i].py;
    if (p[i].py>maxy) maxy=p[i].py;
  }
  if (mapping==1) shade=(int)(128-(p[0].x+p[1].x+p[2].x+p[3].x)*32);
  mapfunc24[mapping](s);
}










void mapside32_noshading(int s)
{
  int x,y,xn;
  unsigned int a,b;
  unsigned da[2],db1[2];
  unsigned char *t=texture[s];
  unsigned long *bf;
  double A0,B0,C0;
  double cc=65536;

  for (y=miny;y<maxy;y++) {
    int v=scr_oy-y;
    int x0=lptr[y];
    xn=rptr[y]-x0;
    if (xn<=0) continue;
    int u=x0-scr_ox;
    C0=u*C1+v*C2+VY*SCALE*C3;
    double q=cc/C0;
    A0=u*A1+v*A2+VY*SCALE*A3;
    B0=u*B1+v*B2+VY*SCALE*B3;
    a=(int)(A0*q+BMPC*65536);
    b=(int)(B0*q+BMPC*65536);
    bf=buffer32+y*scr_cols+x0;
    double C2=C1*8;
    double DA=A1*C0-A0*C1;
    double DB1=B1*C0-B0*C1;
    _asm {
      fld     cc
      fld     DA
      fld     DB1
      fld     C2
      fld     C0
      fld     st
      fadd    st,st(2)
      fld     st
      fmul    st,st(2)
      fdivr   st,st(6)
    }
    for (x=0;x<xn;x++) {
      if ((x&7)==0) {
        _asm { 
          fld     st
          fmul    st,st(6)
          fistp   qword ptr da
          fmul    st,st(4)
          fistp   qword ptr db1
          fst     st(1)
          fadd    st,st(2)
          fld     st
          fmul    st,st(2)
          fdivr   st,st(6)
        }
      }
      unsigned int aa=a/65536;
      unsigned int bb=b/65536;
      aa%=BMPA;
      bb%=BMPA;
      *bf=col3[*(t+aa*BMPA+bb)];
      bf++;
      a+=da[0];
      b+=db1[0];
    }
    _asm {
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
    }
  }
}

void mapside32_flat(int s)
{
  int x,y,xn;
  unsigned int a,b;
  unsigned da[2],db1[2];
  unsigned char *t=texture[s];
  unsigned long *bf;
  long *pal;
  double A0,B0,C0;
  double cc=65536;

  pal=paltbl32[shade];
  for (y=miny;y<maxy;y++) {
    int v=scr_oy-y;
    int x0=lptr[y];
    xn=rptr[y]-x0;
    if (xn<=0) continue;
    int u=x0-scr_ox;
    C0=u*C1+v*C2+VY*SCALE*C3;
    double q=cc/C0;
    A0=u*A1+v*A2+VY*SCALE*A3;
    B0=u*B1+v*B2+VY*SCALE*B3;
    a=(int)(A0*q+BMPC*65536);
    b=(int)(B0*q+BMPC*65536);
    bf=buffer32+y*scr_cols+x0;
    double C2=C1*8;
    double DA=A1*C0-A0*C1;
    double DB1=B1*C0-B0*C1;
    _asm {
      fld     cc
      fld     DA
      fld     DB1
      fld     C2
      fld     C0
      fld     st
      fadd    st,st(2)
      fld     st
      fmul    st,st(2)
      fdivr   st,st(6)
    }
    for (x=0;x<xn;x++) {
      if ((x&7)==0) {
        _asm { 
          fld     st
          fmul    st,st(6)
          fistp   qword ptr da
          fmul    st,st(4)
          fistp   qword ptr db1
          fst     st(1)
          fadd    st,st(2)
          fld     st
          fmul    st,st(2)
          fdivr   st,st(6)
        }
      }
      unsigned int aa=a/65536;
      unsigned int bb=b/65536;
      aa%=BMPA;
      bb%=BMPA;
      *bf=pal[*(t+aa*BMPA+bb)];
      bf++;
      a+=da[0];
      b+=db1[0];
    }
    _asm {
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
    }
  }
}

void mapside32_goraud(int s)
{
  int x,y,xn;
  unsigned int a,b;
  unsigned da[2],db1[2];
  unsigned char *t=texture[s];
  unsigned long *bf;
  double A0,B0,C0;
  double cc=65536;

  for (y=miny;y<maxy;y++) {
    int v=scr_oy-y;
    int x0=lptr[y];
    xn=rptr[y]-x0;
    if (xn<=0) continue;
    int u=x0-scr_ox;
    C0=u*C1+v*C2+VY*SCALE*C3;
    double q=cc/C0;
    A0=u*A1+v*A2+VY*SCALE*A3;
    B0=u*B1+v*B2+VY*SCALE*B3;
    a=(int)(A0*q+BMPC*65536);
    b=(int)(B0*q+BMPC*65536);
    unsigned int sx=(128+lx[y])*65536;
    unsigned int dsx=(rx[y]-lx[y])*65536/xn;
    bf=buffer32+y*scr_cols+x0;
    double C2=C1*8;
    double DA=A1*C0-A0*C1;
    double DB1=B1*C0-B0*C1;
    _asm {
      fld     cc
      fld     DA
      fld     DB1
      fld     C2
      fld     C0
      fld     st
      fadd    st,st(2)
      fld     st
      fmul    st,st(2)
      fdivr   st,st(6)
    }
    for (x=0;x<xn;x++) {
      if ((x&7)==0) {
        _asm { 
          fld     st
          fmul    st,st(6)
          fistp   qword ptr da
          fmul    st,st(4)
          fistp   qword ptr db1
          fst     st(1)
          fadd    st,st(2)
          fld     st
          fmul    st,st(2)
          fdivr   st,st(6)
        }
      }
      unsigned int aa=a/65536;
      unsigned int bb=b/65536;
      aa%=BMPA;
      bb%=BMPA;
      unsigned int ssx=sx/65536;
      *bf=paltbl32[ssx][*(t+aa*BMPA+bb)];
      bf++;
      a+=da[0];
      b+=db1[0];
      sx+=dsx;
    }
    _asm {
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
    }
  }
}

void mapside32_phong(int s)
{
  int x,y,xn;
  unsigned int a,b;
  unsigned da[2],db1[2];
  unsigned char *t=texture[s];
  unsigned long *bf;
  double A0,B0,C0;
  static double cc=65536;

  for (y=miny;y<maxy;y++) {
    int v=scr_oy-y;
    int x0=lptr[y];
    xn=rptr[y]-x0;
    if (xn<=0) continue;
    int u=x0-scr_ox;
    C0=u*C1+v*C2+VY*SCALE*C3;
    double q=cc/C0;
    A0=u*A1+v*A2+VY*SCALE*A3;
    B0=u*B1+v*B2+VY*SCALE*B3;
    a=(int)(A0*q+BMPC*65536);
    b=(int)(B0*q+BMPC*65536);
    unsigned int sa=(int)(la[y]+128)*65536;
    unsigned int dsa=(int)(ra[y]-la[y])*65536/xn;
    unsigned int sb=(int)(lb[y]+128)*65536;
    unsigned int dsb=(int)(rb[y]-lb[y])*65536/xn;
    bf=buffer32+y*scr_cols+x0;
    double C2=C1*8;
    double DA=A1*C0-A0*C1;
    double DB1=B1*C0-B0*C1;
    _asm {
      fld     cc
      fld     DA
      fld     DB1
      fld     C2
      fld     C0
      fld     st
      fadd    st,st(2)
      fld     st
      fmul    st,st(2)
      fdivr   st,st(6)
    }
    for (x=0;x<xn;x++) {
      if ((x&7)==0) {
        _asm { 
          fld     st
          fmul    st,st(6)
          fistp   qword ptr da
          fmul    st,st(4)
          fistp   qword ptr db1
          fst     st(1)
          fadd    st,st(2)
          fld     st
          fmul    st,st(2)
          fdivr   st,st(6)
        }
      }
      unsigned char ssa=sa/65536;
      unsigned char ssb=sb/65536;
      unsigned char ss=phongtbl[ssa][ssb];
      unsigned int aa=a/65536;
      unsigned int bb=b/65536;
      aa%=BMPA;
      bb%=BMPA;
      *bf=paltbl32[ss][*(t+aa*BMPA+bb)];
      bf++;
      a+=da[0];
      b+=db1[0];
      sa+=dsa;
      sb+=dsb;
    }
    _asm {
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
      fstp    st
    }
  }
}

void (*mapfunc32[])(int s)={mapside32_noshading,mapside32_flat,
                          mapside32_goraud,mapside32_phong};

void drawside32(int s)
{
  Tpoint p[4];
  int *ptr1,*ptr2,*sx1,*sx2;
  signed char *sa1,*sa2,*sb1,*sb2;

  if (mapping>3) return;
  p[0]=point[side[s].p1];
  p[1]=point[side[s].p2];
  p[2]=point[side[s].p3];
  p[3]=point[side[s].p4];
  double dx1=p[1].px-p[0].px;
  double dy1=p[1].py-p[0].py;
  double dx2=p[3].px-p[0].px;
  double dy2=p[3].py-p[0].py;
  if (dx1*dy2-dx2*dy1>0) {
    ptr1=rptr;
    sx1=rx;
    sa1=ra;
    sb1=rb;
    ptr2=lptr;
    sx2=lx;
    sa2=la;
    sb2=lb;
  }
  else return;
  A1=p[3].z*p[0].y-p[3].y*p[0].z;
  A2=p[3].y*p[0].x-p[3].x*p[0].y;
  A3=p[3].x*p[0].z-p[3].z*p[0].x;

  B1=p[1].y*p[0].z-p[1].z*p[0].y;
  B2=p[1].x*p[0].y-p[1].y*p[0].x;
  B3=p[1].z*p[0].x-p[1].x*p[0].z;

  C1=(p[3].y*p[1].z-p[3].z*p[1].y+B1+A1)/(BMPA-BMPB);
  C2=(p[3].x*p[1].y-p[3].y*p[1].x+B2+A2)/(BMPA-BMPB);
  C3=(p[3].z*p[1].x-p[3].x*p[1].z+B3+A3)/(BMPA-BMPB);
  miny=scr_rows;maxy=0;
  for (int i=0;i<4;i++) {
    int j=(i+1)&3;
    if (p[i].py<p[j].py) {
      setptr(ptr1,p[i].px,p[i].py,p[j].px,p[j].py);
      if (mapping==2) setptr(sx1,-73*p[i].x,p[i].py,-73*p[j].x,p[j].py); 
      if (mapping==3) {
        setptrp(sa1,p[i].a,p[i].py,p[j].a,p[j].py); 
        setptrp(sb1,p[i].b,p[i].py,p[j].b,p[j].py); 
      }
    }
    else {
      setptr(ptr2,p[j].px,p[j].py,p[i].px,p[i].py);
      if (mapping==2) setptr(sx2,-73*p[j].x,p[j].py,-73*p[i].x,p[i].py);
      if (mapping==3) {
        setptrp(sa2,p[j].a,p[j].py,p[i].a,p[i].py);
        setptrp(sb2,p[j].b,p[j].py,p[i].b,p[i].py);
      }
    }
    if (p[i].py<miny) miny=p[i].py;
    if (p[i].py>maxy) maxy=p[i].py;
  }
  if (mapping==1) shade=(int)(128-(p[0].x+p[1].x+p[2].x+p[3].x)*32);
  mapfunc32[mapping](s);
}

void (*drawfunc[])(int s)={drawside8,drawside16,drawside24,drawside32};

void Tcubeapp::draw( void )
{
  char s[256];

  project();
  buffer8=(unsigned char *)buffer;
  buffer16=(unsigned short *)buffer;
  buffer24=(unsigned char *)buffer;
  buffer32=(unsigned long *)buffer;
  for (int i=0;i<SIDEN;i++)
    drawfunc[bits](i);
  sprintf(s,"%d FPS",FrameRate);
  textout(0,10,s);
  rotx(ax);roty(ay);rotz(az);
}

void Tcubeapp::eventhandler(Tevent *ev)
{
  if (ev->type==evKEY) {
    if (ev->data&128) {
      switch(ev->data&127) {
        case 72: if (ax<MAXA) ax+=A; break;
        case 80: if (ax>-MAXA) ax-=A; break;
        case 75: if (az<MAXA) az+=A; break;
        case 77: if (az>-MAXA) az-=A; break;
        case 71: if (ay<MAXA) ay+=A; break;
        case 73: if (ay>-MAXA) ay-=A; break;
        case 76: ax=ay=az=0; break;
        case  2: mapping=0; break;
        case  3: mapping=1; break;
        case  4: mapping=2; break;
        case  5: mapping=3; break;
        case  6: mapping=4; break;
        case  7: mapping=5; break;
      }
    }
  }
}

#ifdef __WINDOWS__
int Tcubeapp::wndproc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
  HMENU hm;

  switch (msg) {
    case WM_COMMAND:
      switch(LOWORD(wParam)) {
        case ID_OPTIONS_NOSHADING:
          hm=GetSubMenu(GetMenu(hwnd),1);
          if (hm==NULL) break;
          CheckMenuItem(hm,mapping,MF_UNCHECKED|MF_BYPOSITION);
          mapping=0;
          CheckMenuItem(hm,mapping,MF_CHECKED|MF_BYPOSITION);
          break;
        case ID_OPTIONS_FLATSHADING:
          hm=GetSubMenu(GetMenu(hwnd),1);
          CheckMenuItem(hm,mapping,MF_UNCHECKED|MF_BYPOSITION);
          mapping=1;
          CheckMenuItem(hm,mapping,MF_CHECKED|MF_BYPOSITION);
          break;
        case ID_OPTIONS_GORAUDSHADING:
          hm=GetSubMenu(GetMenu(hwnd),1);
          CheckMenuItem(hm,mapping,MF_UNCHECKED|MF_BYPOSITION);
          mapping=2;
          CheckMenuItem(hm,mapping,MF_CHECKED|MF_BYPOSITION);
          break;
        case ID_OPTIONS_PHONGSHADING:
          hm=GetSubMenu(GetMenu(hwnd),1);
          CheckMenuItem(hm,mapping,MF_UNCHECKED|MF_BYPOSITION);
          mapping=3;
          CheckMenuItem(hm,mapping,MF_CHECKED|MF_BYPOSITION);
          break;
        case ID_OPTIONS_TRANSPARENCE:
          hm=GetSubMenu(GetMenu(hwnd),1);
          CheckMenuItem(hm,mapping,MF_UNCHECKED|MF_BYPOSITION);
          mapping=4;
          CheckMenuItem(hm,mapping,MF_CHECKED|MF_BYPOSITION);
          break;
        case ID_OPTIONS_INCLINEDMAPPING:
          hm=GetSubMenu(GetMenu(hwnd),1);
          CheckMenuItem(hm,mapping,MF_UNCHECKED|MF_BYPOSITION);
          mapping=5;
          CheckMenuItem(hm,mapping,MF_CHECKED|MF_BYPOSITION);
          break;
      }
      return 0L;
  }
  return DefWindowProc(hwnd,msg,wParam,lParam);
}
#endif

int Tcubeapp::setmode(int width,int height,int bpp)
{
  int i,j;

  if (!TGRapp::setmode(width,height,bpp)) return 0;
  SCALE=width/10;
  if (bpp==8) bits=0;
  if (bpp==16) {
    bits=1;
    for (i=0;i<256;i++)
      col2[i]=GRrgb(col[i].r,col[i].g,col[i].b);
    for (i=0;i<256;i++)
      for (j=0;j<256;j++) {
        int r=col[j].r*i/200;
        if (r<0) r=0;
        if (r>255) r=255;
        int g=col[j].g*i/200;
        if (g<0) g=0;
        if (g>255) g=255;
        int b=col[j].b*i/200;
        if (b<0) b=0;
        if (b>255) b=255;
        paltbl16[i][j]=GRrgb(r,g,b);
    }
    for (i=0;i<256;i++)
      for (j=0;j<256;j++) {
        if (i) {
          unsigned char cr,cg,cb;
          GRtorgb(i,&cr,&cg,&cb);
          int r=(col[j].r+cr)/2;
          int g=(col[j].g+cg)/2;
          int b=(col[j].b+cb)/2;
          trtbl1[i][j]=GRrgb(r,g,b);
        }
        else trtbl1[i][j]=0;
      }
    for (i=0;i<256;i++)
      for (j=0;j<256;j++) {
        if (i) {
          unsigned char cr,cg,cb;
          GRtorgb(i*256,&cr,&cg,&cb);
          int r=(col[j].r+cr)/2;
          int g=(col[j].g+cg)/2;
          int b=(col[j].b+cb)/2;
          trtbl2[i][j]=GRrgb(r,g,b)&(~127);
        }
        else trtbl2[i][j]=GRrgb(col[j].r,col[j].g,col[j].b);
      }
  }
  if (bpp==24) bits=2;
  if (bpp==32) bits=3;

  if (lptr!=NULL) free(lptr);
  if (rptr!=NULL) free(rptr);
  if (lx!=NULL) free(lx);
  if (rx!=NULL) free(rx);
  if (la!=NULL) free(la);
  if (ra!=NULL) free(ra);
  if (lb!=NULL) free(lb);
  if (rb!=NULL) free(rb);
  if (dy!=NULL) free(dy);
  if (ddy!=NULL) free(ddy);
  lptr=(int *)malloc(sizeof(int)*scr_cols);
  rptr=(int *)malloc(sizeof(int)*scr_cols);
  lx=(int *)malloc(sizeof(int)*scr_cols);
  rx=(int *)malloc(sizeof(int)*scr_cols);
  la=(signed char *)malloc(scr_cols);
  ra=(signed char *)malloc(scr_cols);
  lb=(signed char *)malloc(scr_cols);
  rb=(signed char *)malloc(scr_cols);
  dy=(int *)malloc(sizeof(int)*scr_cols);
  ddy=(int *)malloc(sizeof(int)*scr_cols);
  if (lptr==NULL || rptr==NULL) return 0;
  if (lx==NULL || rx==NULL) return 0;
  if (la==NULL || ra==NULL) return 0;
  if (lb==NULL || rb==NULL) return 0;
  if (dy==NULL || ddy==NULL) return 0;
  mapping=0;
  return 1;
}

int Tcubeapp::init( void )
{
  FILE *f=fopen("CUBE.PAL","rb");
  if (f==NULL) return 0;
  fread(col,3,256,f);
  fclose(f);
  f=fopen("PALTBL1.DAT","rb");
  if (f==NULL) return 0;
  fread(paltbl8,256,256,f);
  fclose(f);
  f=fopen("PALTBL2.DAT","rb");
  if (f==NULL) return 0;
  fread(trtbl8,256,256,f);
  fclose(f);
  for (int a=-128;a<128;a++)
    for (int b=-128;b<128;b++) {
      double x=cos(a*PI/128)*cos(b*PI/256);
      double y=sin(a*PI/128)*cos(b*PI/256);
      double z=sin(b*PI/256);
      double d=pow((x*0.25-y*0.96+1)/2,10);
      int p=(int)(256*d);
      phongtbl[a+128][b+128]=p;
    }
  for (int i=0;i<256;i++)
    col3[i]=col[i].b+(col[i].g<<8)+(col[i].r<<16);
  for (i=0;i<256;i++)
    for (int j=0;j<256;j++) {
      int r=col[j].r*i/200;
      if (r<0) r=0;
      if (r>255) r=255;
      int g=col[j].g*i/200;
      if (g<0) g=0;
      if (g>255) g=255;
      int b=col[j].b*i/200;
      if (b<0) b=0;
      if (b>255) b=255;
      paltbl32[i][j]=b+(g<<8)+(r<<16);
    }
  ax=3*A;ay=A;az=3*A;
  if (!readtextures()) return 0;
  initcube();
  palette=col;
  return setmode(640,480,8);
}

void Tcubeapp::tini( void )
{
  for (int i=0;i<6;i++)
    if (texture[i]!=NULL) free(texture[i]);
  if (lptr!=NULL) free(lptr);
  if (rptr!=NULL) free(rptr);
  if (lx!=NULL) free(lx);
  if (rx!=NULL) free(rx);
  if (la!=NULL) free(la);
  if (ra!=NULL) free(ra);
  if (lb!=NULL) free(lb);
  if (rb!=NULL) free(rb);
  if (dy!=NULL) free(dy);
  if (ddy!=NULL) free(ddy);
}
