#define INITGUID

#include <windows.h>
#include <windowsx.h>
#include <mmsystem.h>
#include <ddraw.h>
#include <dinput.h>

#define _GRLIB_CPP
#include "grlib.h"
#include "resource.h"

int FrameCount;
int FrameCount0;
DWORD FrameTime;
DWORD FrameTime0;

static HINSTANCE hInstApp;
static int       fAppActive;
static int       fAppPaused;
static HWND      hwndApp;
static HACCEL    hAccelApp;

IDirectDraw        *ddraw;
IDirectDrawSurface *FrontBuffer;
IDirectDrawSurface *BackBuffer;
IDirectDrawPalette *Palette;

#define MAXKEYS 10

IDirectInput *din;
IDirectInputDevice *kbd;
DIDEVICEOBJECTDATA didata[MAXKEYS];

TGRapp::TGRapp(char *_title)
{
  title=_title;
  grapp=this;
}

TGRapp::~TGRapp( void )
{
}

int DDInit( void )
{
  DIPROPDWORD dipw;

  if (DirectDrawCreate(NULL,&ddraw,NULL)!=DD_OK) return 0;
  if (ddraw->SetCooperativeLevel(hwndApp, DDSCL_ALLOWREBOOT |
      DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN | DDSCL_ALLOWMODEX)!=DD_OK) return 0;
  if (DirectInputCreate(hInstApp,DIRECTINPUT_VERSION,&din,NULL)!=DI_OK) return 0;
  if (din->CreateDevice(GUID_SysKeyboard,&kbd,NULL)!=DI_OK) return 0;
  if (kbd->SetDataFormat(&c_dfDIKeyboard)!=DI_OK) return 0;
  if (kbd->SetCooperativeLevel(hwndApp,DISCL_FOREGROUND|DISCL_NONEXCLUSIVE)!=DI_OK) return 0;
  dipw.diph.dwSize=sizeof(DIPROPDWORD);
  dipw.diph.dwHeaderSize=sizeof(DIPROPHEADER);
  dipw.diph.dwObj=0;
  dipw.diph.dwHow=DIPH_DEVICE;
  dipw.dwData=MAXKEYS;
  if (kbd->SetProperty(DIPROP_BUFFERSIZE,&dipw.diph)!=DI_OK) return 0;
  kbd->Acquire();
  return 1;
}

int TGRapp::setmode(int width,int height,int bpp)
{
  HRESULT err;

  if (ddraw->SetDisplayMode(width,height,bpp)!=DD_OK) return 0;
  scr_cols=width;
  scr_rows=height;
  scr_bpp=bpp;
  scr_ox=width/2;
  scr_oy=height/2;
  if (BackBuffer)  BackBuffer->Release(),  BackBuffer = NULL;
  if (FrontBuffer) FrontBuffer->Release(), FrontBuffer = NULL;
  if (Palette)     Palette->Release(),     Palette = NULL;

  //
  // Create surfaces
  //
  // what we want is a tripple buffered surface in video memory
  // so we try to create this first.
  //
  // if we cant get a triple buffered surface, we try again
  // for a double buffered surface (still must be in video memory)
  //
  // if we cant get a double buffered surface, we try for a double
  // buffered surface not being specific about video memory, we will
  // get back a main-memory surface, that work use HW page flipping
  // but at least we run.
  //
  // NOTE you need to recreate the surfaces for a new display mode
  // they wont work when/if the mode is changed.
  //
  DDSURFACEDESC ddsd;

  ZeroMemory(&ddsd,sizeof(ddsd));
  ddsd.dwSize=sizeof(ddsd);
  ddsd.dwFlags=DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
  ddsd.dwBackBufferCount=2;
  ddsd.ddsCaps.dwCaps=DDSCAPS_PRIMARYSURFACE |
                      DDSCAPS_FLIP |
                      DDSCAPS_COMPLEX |
                      DDSCAPS_VIDEOMEMORY;

  // try to get a triple buffered video memory surface.
  err=ddraw->CreateSurface(&ddsd,&FrontBuffer,NULL);
  if (err!=DD_OK) {
    // try to get a double buffered video memory surface.
    ddsd.dwBackBufferCount = 1;
    err = ddraw->CreateSurface(&ddsd, &FrontBuffer, NULL);
  }

  if (err!=DD_OK) {
    // settle for a main memory surface.
    ddsd.ddsCaps.dwCaps &= ~DDSCAPS_VIDEOMEMORY;
    err = ddraw->CreateSurface(&ddsd, &FrontBuffer, NULL);
  }

  if (err!=DD_OK) return 0;

  // get a pointer to the back buffer
  DDSCAPS caps;
  caps.dwCaps=DDSCAPS_BACKBUFFER;
  if (FrontBuffer->GetAttachedSurface(&caps,&BackBuffer)!=DD_OK) return 0;

  // create a palette if we are in a paletized display mode.
  PALETTEENTRY ape[256];

  HDC hdc=GetDC(NULL);
  if (GetDeviceCaps(hdc, RASTERCAPS) & RC_PALETTE) {
    GetSystemPaletteEntries(hdc, 0, 256, ape);
    if (palette!=NULL) {
      for (int i=10; i<255; i++) {
        ape[i].peRed  =palette[i].r;
        ape[i].peGreen=palette[i].g;
        ape[i].peBlue =palette[i].b;
      }
    }
    if (ddraw->CreatePalette(DDPCAPS_8BIT,ape,&Palette,NULL)==DD_OK)
      FrontBuffer->SetPalette(Palette);
    GetSystemPaletteEntries(hdc, 0, 256, ape);
  }
  ReleaseDC(NULL, hdc);
  return 1;
}

void DDTerm()
{
  if (BackBuffer)  BackBuffer->Release(),  BackBuffer = NULL;
  if (FrontBuffer) FrontBuffer->Release(), FrontBuffer = NULL;
  if (Palette)     Palette->Release(),     Palette = NULL;
  if (ddraw)       ddraw->Release(),       ddraw = NULL;
  if (kbd) {
    kbd->Unacquire();
    kbd->Release();
    kbd=NULL;
  }
  if (din) din->Release(),din=NULL;
}

HRESULT CALLBACK ModeCallback(LPDDSURFACEDESC pdds, LPVOID lParam)
{
  HMENU hmenu=(HMENU)lParam;
  char ach[80];
  int n;
  int width =pdds->dwWidth;
  int height=pdds->dwHeight;
  int bpp   =pdds->ddpfPixelFormat.dwRGBBitCount;

  if (bpp==8 || bpp==16 || bpp==24 || bpp==32) {
    n=GetMenuItemCount(hmenu);
    wsprintf(ach,"%dx%dx%d",width,height,bpp);
    AppendMenu(hmenu,MF_STRING,MENU_MODE+n,ach);

    MENUITEMINFO mii;

    // pack the mode info into a DWORD and set the extra item data.
    mii.cbSize=sizeof(mii);
    mii.fMask=MIIM_DATA;
    mii.dwItemData=width | (height<<12) | (bpp<<24);
    SetMenuItemInfo(hmenu,MENU_MODE+n,MF_BYCOMMAND,&mii);
  }
  return S_FALSE;
}

void AppPause(int f)
{
  if (f) {
    DDSCAPS caps;
    FrontBuffer->GetCaps(&caps);
    if (caps.dwCaps & DDSCAPS_MODEX) grapp->setmode(640,480,8);
    fAppPaused = 1;
    ddraw->FlipToGDISurface();
    DrawMenuBar(hwndApp);
    RedrawWindow(hwndApp, NULL, NULL, RDW_FRAME);
  }
  else fAppPaused=0;
}

void TGRapp::oneframe( void )
{
  if (FrontBuffer->IsLost()==DDERR_SURFACELOST) FrontBuffer->Restore();
  DDBLTFX ddbltfx;
  ddbltfx.dwSize = sizeof(ddbltfx);
  ddbltfx.dwFillColor = 0;
  BackBuffer->Blt(NULL,NULL,NULL,DDBLT_COLORFILL | DDBLT_WAIT,&ddbltfx);
  DDSURFACEDESC ddsd;

  ddsd.dwSize=sizeof(ddsd);
  if (BackBuffer->Lock(NULL,&ddsd,DDLOCK_WAIT,NULL)!=DD_OK) return;
  buffer=(char *)ddsd.lpSurface;
  BackBuffer->Unlock(NULL);
  draw();

  FrontBuffer->Flip(NULL,DDFLIP_WAIT);

  FrameCount++;
  FrameTime=timeGetTime();
  if (FrameTime-FrameTime0>1000) {
    FrameRate=(FrameCount-FrameCount0)*1000/(FrameTime-FrameTime0);
    FrameTime0=FrameTime;
    FrameCount0=FrameCount;
  }
}

int AppIdle( void )
{
  DWORD i;
  Tevent ev;
  HRESULT err;

  if (fAppActive && !fAppPaused) {
    //*** If the app is active, spin the cube and redraw
    grapp->oneframe();
    err=DIERR_INPUTLOST;
    while (err==DIERR_INPUTLOST) {
      i=MAXKEYS;
      err=kbd->GetDeviceData(sizeof(DIDEVICEOBJECTDATA),didata,&i,0);
      if (err==DIERR_INPUTLOST && kbd->Acquire()!=DI_OK) break;
    }
    if (err!=DI_OK) return 0;
    for (DWORD j=0;j<i;j++) {
      ev.type=evKEY;
      ev.time=didata[j].dwTimeStamp;
      ev.data=(didata[j].dwOfs&127)|((didata[j].dwOfs&128)<<1)|(didata[j].dwData&128);
      grapp->eventhandler(&ev);
    }
    return 0;
  }
  else {
    //*** Don't do anything when not the active app
    return 1;
  }
}

void TGRapp::textout(int x,int y,char *s)
{
  HDC hdc;

  if (BackBuffer->GetDC(&hdc)==DD_OK) {
    SetBkMode(hdc,TRANSPARENT);
    SetTextColor(hdc,RGB(255,255,255));
    TextOut(hdc,x,y,s,strlen(s));
    BackBuffer->ReleaseDC(hdc);
  }
}

int GRrgb(int r,int g,int b)
{
  return ((b>>3)<<0)+
         ((g>>3)<<5)+
         ((r>>3)<<10);
}

void GRtorgb(int col,unsigned char *r,unsigned char *g,unsigned char *b)
{
  *b=(unsigned char)((col>>0)<<3);
  *g=(unsigned char)((col>>5)<<3);
  *r=(unsigned char)((col>>10)<<3);
}

BOOL FAR PASCAL AppAbout(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
  switch (msg) {
    case WM_COMMAND:
      if (LOWORD(wParam)==IDOK) EndDialog(hwnd,TRUE);
      break;
    case WM_INITDIALOG:
      return TRUE;
  }
  return FALSE;
}

LONG CALLBACK AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
  RECT Rect;

  switch (msg) {
    case WM_ACTIVATEAPP:
      fAppActive = (int)wParam;
      break;

    case WM_SETCURSOR:
      if (fAppActive && !fAppPaused) {
        SetCursor(NULL);
        return 1;
      }
      break;

    case WM_LBUTTONDOWN:
      AppPause(!fAppPaused);
      break;

    case WM_ENTERMENULOOP:
      AppPause(1);
      break;

    case WM_EXITMENULOOP:
      AppPause(0);
      break;
   
    case WM_COMMAND:
      switch(LOWORD(wParam)) {
        case MENU_ABOUT:
          AppPause(1);
          DialogBox(hInstApp,"AppAbout",hwnd,(DLGPROC)AppAbout);
          AppPause(0);
          break;

        case MENU_EXIT:
          PostMessage(hwnd,WM_CLOSE,0,0L);
          break;
      }
      if (LOWORD(wParam)>=MENU_MODE && LOWORD(wParam)<MENU_MODE+100) {
        MENUITEMINFO mii;

        mii.cbSize=sizeof(mii);
        mii.fMask=MIIM_DATA;
        GetMenuItemInfo(GetMenu(hwnd),LOWORD(wParam),MF_BYCOMMAND, &mii);

        grapp->setmode(
            (mii.dwItemData >> 0)  & 0xFFF,
            (mii.dwItemData >> 12) & 0xFFF,
            (mii.dwItemData >> 24) & 0x0FF);
        return 0L;
      }
      break;

    case WM_DESTROY:
      hwndApp = NULL;
      PostQuitMessage(0);
      break;

    case WM_MOVE:
    case WM_SIZE:
    case WM_DISPLAYCHANGE:
      if (fAppActive && !IsIconic(hwnd)) {
        SetRect(&Rect, 0, GetSystemMetrics(SM_CYCAPTION), GetSystemMetrics(SM_CXSCREEN), GetSystemMetrics(SM_CYSCREEN));
        AdjustWindowRectEx(&Rect, WS_POPUP | WS_CAPTION, 0, 0);
        SetWindowPos(hwnd, NULL, Rect.left, Rect.top, Rect.right-Rect.left, Rect.bottom-Rect.top, SWP_NOACTIVATE | SWP_NOZORDER);
      }
      break;
  }
  return grapp->wndproc(hwnd,msg,wParam,lParam);
}

int AppInit(HINSTANCE hInst,HINSTANCE hPrev,LPSTR szCmdLine,int sw)
{
  WNDCLASS cls;

  /* Save instance handle for DialogBoxes */
  hInstApp=hInst;

  if (!hPrev) {
    //***  Register a class for the main application window
    cls.hCursor      =LoadCursor(0,IDC_ARROW);
    cls.hIcon        =LoadIcon(hInst, "AppIcon");
    cls.lpszMenuName ="AppMenu";
    cls.lpszClassName=grapp->title;
    cls.hbrBackground=(HBRUSH)GetStockObject(BLACK_BRUSH);
    cls.hInstance    =hInst;
    cls.style        =CS_VREDRAW | CS_HREDRAW;
    cls.lpfnWndProc  =(WNDPROC)AppWndProc;
    cls.cbClsExtra   =0;
    cls.cbWndExtra   =0;

    if (!RegisterClass(&cls)) return 0;
  }

  hAccelApp=LoadAccelerators(hInst,"AppAccel");
  hwndApp=CreateWindowEx(
                  WS_EX_APPWINDOW,
                  grapp->title,           // Class name
                  grapp->title,           // Caption
                  WS_POPUP |
                  WS_SYSMENU |
                  WS_CAPTION,
                  0, 0,                // Position
                  640,480,             // Size
                  0,                   // Parent window (no parent)
                  0,                   // use class menu
                  hInst,               // handle to window instance
                  0                    // no params to pass on
                  );
  ShowWindow(hwndApp,sw);
  UpdateWindow(hwndApp);

  if (!DDInit()) return 0;

  HMENU hmenu=CreatePopupMenu();
  ddraw->EnumDisplayModes(0,NULL,(LPVOID)hmenu,ModeCallback);
  AppendMenu(GetMenu(hwndApp),MF_POPUP,(UINT)hmenu,"Modes");

  return 1;
}

int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrev, LPSTR szCmdLine, int sw)
{
  MSG msg;

  if (!AppInit(hInst,hPrev,szCmdLine,sw)) return 0;
  if (!grapp->init()) return 0;
  while (1) {
    if (PeekMessage(&msg,0,0,0,PM_REMOVE)) {
      if (msg.message==WM_QUIT) break;
      if (!hwndApp || !TranslateAccelerator(hwndApp, hAccelApp, &msg)) {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
      }
    }
    else if (AppIdle()) WaitMessage();
  }
  DDTerm();
  grapp->tini();
  return 1;
}
